import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.io.FileInputStream;
import java.io.BufferedReader;
import java.io.FileReader;

import java.util.*;

import org.apache.hadoop.fs.Path;
import org.apache.hadoop.conf.*;
import org.apache.hadoop.io.*;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.Mapper.Context;
import org.apache.hadoop.mapreduce.Reducer;
import org.apache.hadoop.mapreduce.lib.input.*;
import org.apache.hadoop.mapreduce.lib.output.*;

import org.apache.hadoop.util.*;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.FSDataOutputStream;
import org.apache.hadoop.filecache.DistributedCache;

import org.apache.log4j.Logger;
import org.apache.log4j.Level;
import org.apache.log4j.BasicConfigurator;

import java.nio.ByteBuffer;

public class MRConnectedComponents {                                                                                                        
    static private Logger logger = Logger.getLogger(MRConnectedComponents.class);

    public static final Path INPUT_DIR  = new Path("input");
    public static final Path OUTPUT_DIR = new Path("output");
    public static final Path TEMP_DIR   = new Path("intermediate");

    public static class LongArrayWritable extends ArrayWritable
    {
	public LongArrayWritable() {
	    super(LongWritable.class);
	}
	public LongArrayWritable(LongWritable[] values) {
	    super(LongWritable.class, values);
	}
    }



    enum PagasusStats {
	OPEN_COMPONENTS // pagasus flag stopping criterion
	    };

    // THIS IS FOR UNDIRECTED GRAPHS
    // At the first iteration the input an arc, i.e. a pair of items
    public static class Pegasus_Mapper_first extends Mapper<LongWritable, Text, LongWritable, LongArrayWritable> {
	@Override
	    protected void map(LongWritable key, Text value, Context context) throws IOException, InterruptedException {
	    // commented line
	    if (value.toString().startsWith("#")) return;

	    String [] values = value.toString().split("\\t");
	    LongWritable node  = new LongWritable( Long.parseLong(values[0]) );
	    LongWritable neigh = new LongWritable( Long.parseLong(values[1]) );

	    // logger.info("node: " + node.get() +"\t neigh: " + neigh.get() );
	    
	    // self loop
	    if ( node.equals(neigh) ) {
		LongWritable[] singlet = new LongWritable[1];
		singlet[0] = node;
		context.write(node, new LongArrayWritable(singlet) );
		singlet[0] = new LongWritable(-1 -node.get() );
		context.write(node, new LongArrayWritable(singlet) );
	    } else {
		// send graph to node and to neighbor
		LongWritable[] singlet = new LongWritable[1];
		singlet[0] = neigh;
		context.write(node, new LongArrayWritable(singlet) );
		singlet[0] = node;
		context.write(neigh, new LongArrayWritable(singlet) );

		singlet[0] = new LongWritable(-1 - Math.min(node.get(), neigh.get()));
		context.write(node, new LongArrayWritable(singlet) );
		context.write(neigh, new LongArrayWritable(singlet) );
	    }
	}
    }

    // in successive iterations the input is an array of node ids
    public static class Pegasus_Mapper extends Mapper<LongWritable, LongArrayWritable, LongWritable, LongArrayWritable> {
	// assume nodes are sorted in the longarraywriteable
	@Override
	    protected void map(LongWritable key, LongArrayWritable value, Context context) 
	    throws IOException, InterruptedException {
	    Writable[] min_nodes = value.get();	    
	    LongWritable min = (LongWritable) min_nodes[0];

	    // send min to self
	    LongWritable[] singlet = new LongWritable[1];
	    singlet[0] = min;
	    context.write(key, new LongArrayWritable(singlet));

	    int num_neigh = min_nodes.length -1;
	    if (num_neigh>0) { // at least one outgoing link
		LongWritable[] nodes = new LongWritable[num_neigh];
		for (int i=0; i<nodes.length; i++)
		    nodes[i] = (LongWritable) min_nodes[i+1];

		// send graph
		context.write(key, new LongArrayWritable(nodes));

		// send min all other nodes
		for (int i=0; i<nodes.length; i++) {
		    context.write( nodes[i], new LongArrayWritable(singlet)); 
		}
	    }
	}
    }

    public static class Pegasus_Reducer extends Reducer<LongWritable, LongArrayWritable, LongWritable, LongArrayWritable> {
	@Override
	    protected void reduce(LongWritable key, Iterable<LongArrayWritable> values, Context context) 
	    throws IOException, InterruptedException {
	    // find smallest min and the graph neighbors
	    LongWritable min = null;
	    Writable[] neighbors = null;
	    boolean min_agreement = true;
	    for(LongArrayWritable cluster : values) {
		Writable[] nodes = cluster.get();
		LongWritable first = (LongWritable) nodes[0];
		if (first.get()<0) { // This is a min !
		    if (min==null) min = first;
		    else if (min.compareTo(first)<0) {
			min = first;
			min_agreement = false;
		    }
		} else { // This is the actual graph
		    if (neighbors == null)
			neighbors = nodes;
		    else {
			Writable[] aux = new Writable[neighbors.length + nodes.length];
			System.arraycopy(neighbors, 0, aux, 0, neighbors.length);
			System.arraycopy(nodes, 0, aux, neighbors.length, nodes.length);
			neighbors = aux;
		    }
		}
	    }

	    // number of neighbors
	    int num_neigh = neighbors==null ? 0 : neighbors.length;
	    LongWritable[] min_nodes = new LongWritable[num_neigh + 1];
	    min_nodes[0] = min;
	    for (int i=0; i<num_neigh; i++)
		min_nodes[i+1] = (LongWritable) neighbors[i];

	    // send current min + graph
	    context.write(key, new LongArrayWritable(min_nodes));
	    
	    if (!min_agreement)
		context.getCounter(PagasusStats.OPEN_COMPONENTS).increment(1);
	}
    }

    // this class output clusters in the form <node id, component id>
    public static class Pegasus_Outputter extends Mapper<LongWritable, LongArrayWritable, LongWritable, LongWritable> {
	// assume nodes are sorted in the longarraywriteable
	@Override
	    protected void map(LongWritable key, LongArrayWritable value, Context context) throws IOException, InterruptedException {
	    Writable[] cluster = value.get();
	    LongWritable min = new LongWritable( -1-((LongWritable) cluster[0]).get() );
	    context.write(key, min); 
	}
    }

    public static boolean Pegasus (String[] args, Path input_dir, Path output_dir, Path intermediate_results) 
	throws IOException,InterruptedException,ClassNotFoundException {
	Path iter_input_path = null;
	Path iter_output_path = null;
	int max_iteration = 30;

	// Pegasus Algorithm invocation
	Configuration configuration = new Configuration();
	GenericOptionsParser optionsParser = new GenericOptionsParser(configuration, args);
	FileSystem fs = FileSystem.get(configuration);
	for (int iteration = 0; iteration<=max_iteration; iteration++ ) {
	    configuration.set("io.sort.record.percent", "0.05" );
	    Job jobStep = new Job(configuration, "PEGASUS iteration " + iteration);
	    jobStep.setJarByClass(MRConnectedComponents.class);
	    
	    // common settings
	    iter_input_path = new Path(intermediate_results, "iter"+iteration);
	    iter_output_path = new Path(intermediate_results, "iter"+(iteration+1));
	    Class mapper_class        = Pegasus_Mapper.class;
	    Class reducer_class       = Pegasus_Reducer.class;
	    Class output_class        = LongArrayWritable.class;
	    Class output_format_class = SequenceFileOutputFormat.class;
	    Class input_format_class  = SequenceFileInputFormat.class;

	    // per iteration settings
	    if (iteration==0) {
		iter_input_path     = input_dir;
		mapper_class        = Pegasus_Mapper_first.class;
		input_format_class  = TextInputFormat.class;
	    } else if (iteration == max_iteration) {
		mapper_class        = Pegasus_Outputter.class;
		reducer_class       = null;
		iter_output_path    = output_dir;
		output_format_class = TextOutputFormat.class;
		output_class        = LongWritable.class;
	    } 
	    // common
	    if (fs.exists(iter_output_path)) fs.delete(iter_output_path, true);

	    jobStep.setMapperClass(mapper_class);
	    if (reducer_class!=null) {
		jobStep.setReducerClass(reducer_class);
	    }

	    jobStep.setOutputKeyClass(LongWritable.class);
	    jobStep.setOutputValueClass(output_class);
	    
	    jobStep.setInputFormatClass(input_format_class);
	    jobStep.setOutputFormatClass(output_format_class);

	    FileInputFormat.setInputPaths(jobStep, iter_input_path);
	    FileOutputFormat.setOutputPath(jobStep, iter_output_path);
	    
	    boolean success = jobStep.waitForCompletion(true);
	    if (! success) {
		logger.error("Pegasus Job has failed (Iter Loop).");
		return false;
	    }

	    if (iteration>0) fs.delete(iter_input_path, true);

	    if( iteration != max_iteration && jobStep.getCounters().findCounter(PagasusStats.OPEN_COMPONENTS).getValue() == 0 )
		max_iteration = iteration+1;
	}
	return true;
    }

    public static void main(String[] args) throws Exception {
	Pegasus(args, INPUT_DIR, OUTPUT_DIR, TEMP_DIR);
    }
}
